# -*- coding: utf-8 -*-
"""Created on Tue Jul 18 17:59:45 2017.

@author: erwan


Rovibrational Energy for Vibrating Rotor

Ex: used for CO2 X(1Σ+) state

References
----------





-----------------------------------------------------------------------------
"""


from radis.db.conventions import (
    herzberg_coefficients_rot,
    herzberg_coefficients_rovib,
    herzberg_coefficients_vib,
)
from radis.levels.dunham import Fv, Gv


def EvJ_uncoupled_vibrating_rotor(
    v1, v2, l2, v3, J, coeff_dict, gv1=1, gv2=1, gv3=1, remove_ZPE=True
):
    """Rovibrational energy of an uncoupled vibrating rotor.

    Parameters
    ----------

    v1: int
        vibrational state

    v2: int
        vibrational state

    l2: int
        vibrational state

    v3: int
        vibrational state

    J: int
        rotational state

    coeff_dict: dict
        dictionary of :py:data:`~radis.db.conventions.herzberg_coefficients`,
        which can be numbered for the different vibration modes. Example::

            {'we1': 1333.93,
             'we2': 667.47,
             'we3': 2349.16,
             'wexe1': 2.93,
             'wexe2': -0.38,
             'wexe3': 12.47,
             'Be': 0.39022,
             'De': 1.333e-07,
             'He': 9e-15}

    gv1, gv2, gv3: int
        degeneracies of each vibrational mode. Default ``1, 1, 1``::

            1,2,1 for CO2

    remove_ZPE: boolean
        if ``True``, removes energy of ground state vibrational level (zero-point-energy)

    Returns
    -------

    E: float
        energy of state in cm-1

    References
    ----------

    Klarenaar et al, "Time evolution of vibrational temperatures in a CO 2 glow
    discharge measured with infrared absorption spectroscopy", doi 10.1088/1361-6595/aa902e,
    and the references there in.

    See Also
    --------

    :py:func:`~radis.levels.energies_co2.EvJ_co2`
    """

    coeff_dict = coeff_dict.copy()

    coeffs_vib = {"1": {}, "2": {}, "3": {}}
    coeffs_rot = {}

    # Split coeffs between the different vibration and rotation modes:
    for k, v in coeff_dict.items():
        if k in herzberg_coefficients_rot:
            coeffs_rot[k] = v
        elif k[:-1] in herzberg_coefficients_vib:
            vib_mode = k[-1]
            coeffs_vib[vib_mode][k[:-1]] = v
        elif k in herzberg_coefficients_rovib:
            raise NotImplementedError(
                "Mixed term {0} not implemented for uncoupled rovib model".format(k)
            )
        else:
            raise KeyError("Unexpected coefficient: {0}".format(k))

    coeffs_vib1 = coeffs_vib["1"]
    coeffs_vib2 = coeffs_vib["2"]
    coeffs_vib3 = coeffs_vib["3"]

    # Get rotational coeffs:
    coeffs_rot = {
        k: v for (k, v) in coeff_dict.items() if k in herzberg_coefficients_rot
    }

    # Energies
    G1 = Gv(v1, gv=gv1, **coeffs_vib1)
    G2 = Gv(v2, gv=gv2, **coeffs_vib2)
    G3 = Gv(v3, gv=gv3, **coeffs_vib3)
    G = G1 + G2 + G3
    F = Fv(0, J, **coeffs_rot)  # Uncoupled model: v dependance ignored

    if remove_ZPE:
        ZPE = Gv(0, **coeffs_vib1) + Gv(0, gv=2, **coeffs_vib2) + Gv(0, **coeffs_vib3)
    else:
        ZPE = 0

    return G + F - ZPE  # cm-1


def EvJah_uncoupled_vibrating_rotor(v1, v2, l2, v3, J, coeff_dict, remove_ZPE=True):
    """Return rovibrationalenergies for nu_1, nu_2, nu_3 vibrational modes Each
    energy is a tuple (E_harmonic, E_nonharmonic) to be used for instance in a
    Treanor distribution.

    Defined for CO2 only. See:  Klarenaar et al. [1]_

    Energy of CO2 X(1Σg+) state with an independant harmonic oscillator
    approximation  (CO2-626 = first isotope)

    Note that this is less accurate than using CDSD energies in the (p, J, c, n)
    general denomination

    Parameters
    ----------

    v1: int
        vibrational state

    v2: int
        vibrational state

    l2: int
        vibrational state

    v3: int
        vibrational state

    J: int
        rotational state

    remove_ZPE: boolean
        if ``True``, removes energy of ground state vibrational level (zero-point-energy)

    Returns
    -------

    E: float
        energy of state in cm-1

    References
    ----------

    .. [1] Eq.(10) in Klarenaar et al, "Time evolution of vibrational temperatures in a CO2 glow
        discharge measured with infrared absorption spectroscopy", doi 10.1088/1361-6595/aa902e,
        and the references there in.
    """

    coeff_dict = coeff_dict.copy()

    we1 = coeff_dict.pop("we1")
    we2 = coeff_dict.pop("we2")
    we3 = coeff_dict.pop("we3")

    wexe1 = coeff_dict.pop("wexe1")
    wexe2 = coeff_dict.pop("wexe2")
    wexe3 = coeff_dict.pop("wexe3")

    Be = coeff_dict.pop("Be")
    De = coeff_dict.pop("De")
    He = coeff_dict.pop("He")

    if len(coeff_dict) > 0:
        raise NotImplementedError(
            "Harmonic/anharmonic energy split not defined "
            + "with the following spectroscopic constants: {0}".format(
                list(coeff_dict.keys())
            )
        )

    # Vibrational energy
    G1_h = we1 * v1
    G1_a = -wexe1 * v1 * (v1 - 1)
    # G1 = G1_h + G1_a

    G2_h = we2 * v2
    G2_a = -wexe2 * v2 * (v2 - 1)
    # G2 = G2_h + G2_a

    G3_h = we3 * v3
    G3_a = -wexe3 * v3 * (v3 - 1)
    # G3 = G3_h + G3_a

    # G = G1 + G2 + G3

    # Rotational energy
    Bv = Be  # no coupling terms
    Dv = De
    Hv = He
    F = Bv * J * (J + 1) - Dv * J ** 2 * (J + 1) ** 2 + Hv * J ** 3 * (J + 1) ** 3

    if remove_ZPE:
        # ZPE = 0  # by construction here
        pass
    else:
        raise NotImplementedError

    return (G1_h, G1_a), (G2_h, G2_a), (G3_h, G3_a), F


#    return Te + G + F - ZPE   # cm-1


# %% Test


def _test(*args, **kwargs):
    #    E_CO2X = E_CO2_626_X1SIGg
    from radis.db.utils import get_herzberg_coefficients

    coeffs = get_herzberg_coefficients("CO2", 1, "X1SIGu+")
    E_CO2X = lambda v1, v2, l2, v3, J: EvJ_uncoupled_vibrating_rotor(
        v1, v2, l2, v3, J, coeff_dict=coeffs, gv1=1, gv2=2, gv3=1
    )
    print(
        "CO2((0,0,0,0),J=0) -> CO2((0,0,0,1),J=0) energy: ",
        E_CO2X(0, 0, 0, 1, 0) - E_CO2X(0, 0, 0, 0, 0),
        "cm-1",
    )


if __name__ == "__main__":

    _test()



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 18 20:18:34 2021

@author: erwan


Add a RulerTool to measure spectra

    Based on work from TerranJP
    https://github.com/terranjp/matplotlib-tools


Examples
--------
::
    from radis.tools.plot_tools import add_ruler
    fig = plt.figure()
    add_ruler(fig)

.. image:: https://user-images.githubusercontent.com/16088743/122615292-95a9e080-d088-11eb-9927-bf1187d5a94a.png



"""
import warnings


def add_ruler(fig, wunit="", Iunit="", ax=None):

    import matplotlib.pyplot as plt
    import numpy as np
    from matplotlib.backend_tools import ToolToggleBase
    from matplotlib.widgets import AxesWidget

    # TODO: build it only on demand/ if add_tool called?
    class Ruler(AxesWidget):
        """
        A ruler to measure distances and angles on an axes instance.

        For the ruler to remain responsive you must keep a reference to it.

        Parameters
        ----------
        ax  : the  :class:`matplotlib.axes.Axes` instance
        active : bool, default is True
            Whether the ruler is active or not.
        wunit, Iunit: str
            unit of spectra
        print_text  : bool, default is False
            Whether the length measure string is printed to the console
        useblit : bool, default is False
            If True, use the backend-dependent blitting features for faster
            canvas updates.
        lineprops : dict, default is None
          Dictionary of :class:`matplotlib.lines.Line2D` properties
        markerprops : dict, default is None
          Dictionary of :class:`matplotlib.markers.MarkerStyle` properties
        textprops: dict, default is None
            Dictionary of :class:`matplotlib.text.Text` properties. To reposition the
            textbox you can overide the defaults which position the box in the top left
            corner of the axes.

        Notes
        -----

        Usage:

        1. Hold left click drag and release to draw the ruler in the axes.
          - Hold shift while dragging to lock the ruler to the horizontal axis.
          - Hold control while drawing to lock the ruler to the vertical axis.

        2. Right click one of the markers to move the ruler.

        The keyboard can be used to activate and deactivate the ruler and toggle
        visibility of the line and text:

        'm' : Toggles the ruler on and off.

        'ctl+m' : Toggles the visibility of the ruler and text.

        Example
        -------

        >>> xCoord = np.arange(0, 5, 1)
        >>> yCoord = [0, 1, -3, 5, -3]
        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(111)

        >>> markerprops = dict(marker='o', markersize=5, markeredgecolor='red')
        >>> lineprops = dict(color='red', linewidth=2)

        >>> ax.grid(True)
        >>> ax.plot(xCoord, yCoord)

        >>> ruler = Ruler(ax=ax,
                      useblit=True,
                      markerprops=markerprops,
                      lineprops=lineprops)

        >>> plt.show()

        Typical output :

        .. image:: https://user-images.githubusercontent.com/16088743/122615292-95a9e080-d088-11eb-9927-bf1187d5a94a.png

        """

        def __init__(
            self,
            ax,
            active=True,
            print_text=False,
            useblit=False,
            lineprops=None,
            textprops=None,
            markerprops=None,
            wunit="",
            Iunit="",
        ):
            """
            Add a ruler to *ax*. If ``ruler_active=True``, the ruler will be
            activated when the plot is first created. If ``ruler_unit`` is set the
            string will be appended to the length text annotations.

            """
            AxesWidget.__init__(self, ax)

            self.connect_events()

            self.ax = ax
            self.fig = ax.figure

            self._print_text = print_text
            self._visible = True
            self.active = active
            self.wunit = wunit
            self.Iunit = Iunit

            self.useblit = useblit and self.canvas.supports_blit

            self._mouse1_pressed = False
            self._mouse3_pressed = False
            self._shift_pressed = False
            self._control_pressed = False
            self._y0 = None
            self._x1 = None
            self._y1 = None
            self._line_start_coords = None
            self._line_end_coords = None
            self._ruler_marker = None
            self._background = None
            self._ruler_moving = False
            self._end_a_lock = False
            self._end_b_lock = False
            self._end_c_lock = False
            self._old_marker_a_coords = None
            self._old_marker_c_coords = None
            self._old_mid_coords = None

            if lineprops is None:
                lineprops = {}

            bbox = dict(
                facecolor="white", alpha=0.5, boxstyle="round", edgecolor="0.75"
            )

            used_textprops = dict(
                xy=(0, 1),
                xytext=(10, -10),
                xycoords="axes fraction",
                textcoords="offset points",
                ha="left",
                va="center",
                size=12,
                bbox=bbox,
            )

            x0 = np.nan
            y0 = np.nan

            (self._ruler,) = self.ax.plot([x0, x0], [y0, y0], **lineprops)

            used_markerprops = dict(
                marker="s",
                markersize=3,
                markerfacecolor="white",
                markeredgecolor="black",
                markeredgewidth=0.5,
                picker=True,
                pickradius=5,
                visible=False,
            )

            # If marker or text  props are given as an argument combine with the
            # default marker props. Don't really want to override the entire props
            # if a user only gives one value.

            if markerprops is not None:
                used_markerprops.update(markerprops)

            if textprops is not None:
                used_textprops.update(used_textprops)

            self._axes_text = self.ax.annotate(text="", **used_textprops)
            self.ax.add_artist(self._axes_text)

            (self._marker_a,) = self.ax.plot((x0, y0), **used_markerprops)
            (self._marker_b,) = self.ax.plot((x0, y0), **used_markerprops)
            (self._marker_c,) = self.ax.plot((x0, y0), **used_markerprops)

            self._artists = [
                self._axes_text,
                self._ruler,
                self._marker_a,
                self._marker_b,
                self._marker_c,
            ]

        def connect_events(self):
            """
            Connect all events to the various callbacks
            """
            self.connect_event("button_press_event", self._on_press)
            self.connect_event("button_release_event", self._on_release)
            self.connect_event("motion_notify_event", self._on_move)
            self.connect_event("key_press_event", self._on_key_press)
            self.connect_event("key_release_event", self._on_key_release)

        def ignore(self, event):
            """
            Ignore events if the cursor is out of the axes or the widget is locked
            """
            if not self.canvas.widgetlock.available(self):
                return True
            if event.inaxes != self.ax.axes:
                return True
            if not self.active:
                return True

        def _on_key_press(self, event):
            """
            Handle key press events.

            If shift is pressed the ruler will be constrained to horizontal axis
            If control is pressed the ruler will be constrained to vertical axis
            If m is pressed the ruler will be toggled on and off
            If ctrl+m is pressed the visibility of the ruler will be toggled
            """

            if event.key == "shift":
                self._shift_pressed = True

            if event.key == "control":
                self._control_pressed = True

            if event.key == "m":
                self.toggle_ruler()

            if event.key == "ctrl+m":
                self.toggle_ruler_visibility()

        def _on_key_release(self, event):
            """
            Handle key release event, flip the flags to false.
            """
            if event.key == "shift":
                self._shift_pressed = False

            if event.key == "control":
                self._control_pressed = False

        def toggle_ruler(self):
            """
            Called when the 'm' key is pressed. If ruler is on turn it off, and
            vise versa

            If off, hide it.
            """

            self.active = not self.active

            self.canvas.draw_idle()

        def toggle_ruler_visibility(self):
            """
            Called when the 'ctl+m' key is pressed. If ruler is visible turn it off
            , and vise versa
            """
            if self._visible is True:
                for artist in self._artists:
                    artist.set_visible(False)
                self.active = False
                self._visible = False

            elif self._visible is False:
                for artist in self._artists:
                    artist.set_visible(True)
                self._visible = True

            self.canvas.draw_idle()

        def _on_press(self, event):
            """
            On mouse button press check which button has been pressed and handle
            """
            if self.ignore(event):
                return
            if event.button == 1 and self._mouse3_pressed is False:
                self._handle_button1_press(event)
            elif event.button == 3:
                self._handle_button3_press(event)

        def _handle_button1_press(self, event):
            """
            On button 1 press start drawing the ruler line from the initial
            press position
            """

            self._mouse1_pressed = True
            self._x0 = event.xdata
            self._y0 = event.ydata
            self._marker_a.set_data((event.xdata, event.ydata))
            self._marker_a.set_visible(True)

            if self.useblit:
                self._marker_a.set_data(self._x0, self._y0)
                for artist in self._artists:
                    artist.set_animated(True)
                self.canvas.draw()
                self._background = self.canvas.copy_from_bbox(self.fig.bbox)

        def _handle_button3_press(self, event):
            """
            If button 3 is pressed (right click) check if cursor is at one of the
            ruler markers and the move the ruler accordingly.
            """
            contains_a, _ = self._marker_a.contains(event)
            contains_b, _ = self._marker_b.contains(event)
            contains_c, _ = self._marker_c.contains(event)

            if not (contains_a or contains_b or contains_c):
                return

            self._end_a_lock = contains_a
            self._end_b_lock = contains_b
            self._end_c_lock = contains_c

            line_coords = self._ruler.get_path().vertices
            self._x0 = line_coords[0][0]
            self._y0 = line_coords[0][1]
            self._x1 = line_coords[1][0]
            self._y1 = line_coords[1][1]

            self._old_marker_a_coords = self._marker_a.get_path().vertices
            self._old_marker_c_coords = self._marker_c.get_path().vertices
            self._old_mid_coords = self.midline_coords

        def _on_move(self, event):
            """
            On motion draw the ruler if button 1 is pressed. If one of the markers
            is locked indicating move the ruler according to the locked marker
            """

            if event.inaxes != self.ax.axes:
                return

            # if self._end_a_lock or self._end_b_lock or self._end_c_lock is True:
            #     self._move_ruler(event)

            if self._mouse1_pressed is True:
                self._draw_ruler(event)

        # def _move_ruler(self, event):
        #     """
        #     If one of the markers is locked move the ruler according the selected
        #     marker.
        #     """

        #     # This flag is used to prevent the ruler from clipping when a marker is
        #     # first selected
        #     if self._ruler_moving is False:
        #         if self.useblit:
        #             for artist in self._artists:
        #                 artist.set_animated(True)
        #             self.canvas.draw()
        #             self._background = self.canvas.copy_from_bbox(self.fig.bbox)
        #             self._ruler_moving = True

        #     if self._end_a_lock is True:
        #         # If marker a is locked only move end a.
        #         pos_a = event.xdata, self._x1
        #         pos_b = event.ydata, self._y1
        #         self._marker_a.set_data(event.xdata, event.ydata)
        #         self._ruler.set_data(pos_a, pos_b)
        #         self._set_midline_marker()

        #     if self._end_c_lock is True:
        #         # If marker a is locked only move end c.
        #         pos_a = self._x0, event.xdata
        #         pos_b = self._y0, event.ydata
        #         self._marker_c.set_data(event.xdata, event.ydata)
        #         self._ruler.set_data(pos_a, pos_b)
        #         self._set_midline_marker()

        #     if self._end_b_lock is True:
        #         # If marker b is locked shift the whole ruler.
        #         b_dx = event.xdata - self._old_mid_coords[0]
        #         b_dy = event.ydata - self._old_mid_coords[1]
        #         pos_a = self._x0 + b_dx, self._x1 + b_dx
        #         pos_b = self._y0 + b_dy, self._y1 + b_dy

        #         marker_a_coords = (
        #             self._old_marker_a_coords[0][0] + b_dx,
        #             self._old_marker_a_coords[0][1] + b_dy,
        #         )
        #         marker_c_coords = (
        #             self._old_marker_c_coords[0][0] + b_dx,
        #             self._old_marker_c_coords[0][1] + b_dy,
        #         )

        #         self._ruler.set_data(pos_a, pos_b)
        #         self._marker_a.set_data(marker_a_coords)
        #         self._marker_b.set_data(event.xdata, event.ydata)
        #         self._marker_c.set_data(marker_c_coords)

        #     self._update_text()
        #     self._update_artists()

        # def _set_midline_marker(self):
        #     self._marker_b.set_visible(True)
        #     self._marker_b.set_data(self.midline_coords)

        # @property
        # def midline_coords(self):
        #     pos0, pos1 = self._ruler.get_path().vertices
        #     mid_line_coords = (pos0[0] + pos1[0]) / 2, (pos0[1] + pos1[1]) / 2
        #     return mid_line_coords

        def _draw_ruler(self, event):
            """
            If the left mouse button is pressed and held draw the ruler as the
            mouse is dragged
            """

            self._x1 = event.xdata
            self._y1 = event.ydata

            # If shift is pressed ruler is constrained to horizontal axis
            if self._shift_pressed is True:
                pos_a = self._x0, self._x1
                pos_b = self._y0, self._y0
            # If control is pressed ruler is constrained to vertical axis
            elif self._control_pressed is True:
                pos_a = self._x0, self._x0
                pos_b = self._y0, self._y1
            # Else the ruler follow the mouse cursor
            else:
                pos_a = self._x0, self._x1
                pos_b = self._y0, self._y1

            self._ruler.set_data([pos_a], [pos_b])
            x1 = self._ruler.get_path().vertices[1][0]
            y1 = self._ruler.get_path().vertices[1][1]
            self._marker_c.set_visible(True)
            self._marker_c.set_data(x1, y1)
            # self._set_midline_marker()
            self._update_text()
            self._update_artists()

        def _update_artists(self):
            if self.useblit:
                if self._background is not None:
                    self.canvas.restore_region(self._background)
                else:
                    self._background = self.canvas.copy_from_bbox(self.fig.bbox)

                for artist in self._artists:
                    self.ax.draw_artist(artist)

                self.canvas.blit(self.fig.bbox)
            else:
                self.canvas.draw_idle()

        def _update_text(self):
            detail_string = "{:0.4f} {}; {:0.3f} {}".format(
                self.ruler_dx, self.wunit, self.ruler_dy, self.Iunit
            )

            self._axes_text.set_text(detail_string)
            if self._print_text is True:
                print(detail_string)

        def _on_release(self, event):
            self._mouse1_pressed = False
            self._mouse3_pressed = False
            self._ruler_moving = False
            self._end_a_lock = False
            self._end_b_lock = False
            self._end_c_lock = False

        @property
        def ruler_dx(self):
            pos0, pos1 = self._ruler.get_path().vertices
            return pos1[0] - pos0[0]

        @property
        def ruler_dy(self):
            pos0, pos1 = self._ruler.get_path().vertices
            return pos1[1] - pos0[1]

    # %% Add in toolbars

    class RulerTool(ToolToggleBase):
        """Add a RulerTool to measure spectra

        Based on work from TerranJP
        https://github.com/terranjp/matplotlib-tools
        """

        # keyboard shortcut
        default_keymap = "m"
        description = "Ruler Tool"
        default_toggled = False
        image = "ruler"

        def __init__(self, *args, **kwargs):
            ax = kwargs.get("ax", plt.gca())
            self.ruler = None
            self.ax = ax
            self.wunit = (None,)  # x-axis unit
            self.Iunit = (None,)  # y-axis unit
            super().__init__(*args, **kwargs)

        def enable(self, event):
            """Connect press/release events and lock the canvas."""
            pass

        def disable(self, event):
            """Release the canvas and disconnect press/release events."""
            pass

        def trigger(self, *args, **kwargs):
            """When toolbar button clicked"""
            super().trigger(*args, **kwargs)
            if self.ruler is None:
                self.ruler = Ruler(
                    ax=self.ax,  # TODO UPDATE
                    useblit=True,
                    wunit=self.wunit,
                    Iunit=self.Iunit,
                )  # , markerprops=markerprops, lineprops=lineprops)
            else:
                self.ruler.toggle_ruler()

    # %% Build Class

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        plt.rcParams["toolbar"] = "toolmanager"
        # Filters https://github.com/matplotlib/matplotlib/issues/15284

    if fig.canvas.manager.toolmanager is None:
        from warnings import warn

        warn(
            "Couldn't add Ruler tool (still an experimental feature in RADIS : please report the error !)"
        )
        return

    if "📐" in fig.canvas.manager.toolmanager.tools:
        return

    # Add the custom tools that we created
    fig.canvas.manager.toolmanager.add_tool("📐", RulerTool)

    # Define units
    rulertool = fig.canvas.manager.toolmanager.get_tool("📐")
    rulertool.wunit = wunit
    rulertool.Iunit = Iunit
    if ax is not None:  # add the Ruler only on a specific ax
        rulertool.ax = ax

    # Add it to new group `radis_tools`
    fig.canvas.manager.toolbar.add_tool("📐", "radis_tools")


if __name__ == "__main__":

    # fig = plt.figure()
    # plt.plot([1, 2, 3])

    # add_ruler(fig)
    #%%
    from radis import calc_spectrum

    s = calc_spectrum(
        1900,
        2300,  # cm-1
        molecule="CO",
        isotope="1,2,3",
        pressure=1.01325,  # bar
        Tgas=700,  # K
        mole_fraction=0.1,
        path_length=1,  # cm
        databank="hitran",  # or use 'hitemp'
    )
    s.apply_slit(0.5, "nm")  # simulate an experimental slit
    # line = s.plot('radiance', nfig='same')

    from radis import plot_diff

    plot_diff(s, s.take("radiance") * 1.1, show_ruler=True)



# -*- coding: utf-8 -*-
"""Definition of molecules and list of spectroscopic constants
"""

from .molecules import Molecules, getMolecule


# %% Get list of supported molecules
def _get_supported_molecules_equilibrium():
    """Molecules supported in RADIS equilibrium calculations Basically, all
    [HITRAN-2016]_ species."""
    # Add all HITRAN species
    # Add ExoMol species
    from .classes import EXOMOL_MOLECULES, HITRAN_MOLECULES

    return list(set(HITRAN_MOLECULES).union(set(EXOMOL_MOLECULES)))


def _get_supported_molecules_nonequilibrium():
    """Molecules supported in RADIS non equilibrium calculations without need
    for extra databases.

    Basically, molecules whose spectroscopic constants are built-in
    RADIS database (see radis.db)
    """

    return list(Molecules.keys())


MOLECULES_LIST_EQUILIBRIUM = (
    __supported_molecules_equilibrium__
) = _get_supported_molecules_equilibrium()
""" list: molecules that can be calculated in RADIS at equilibrium.
All [HITRAN-2016]_ and [ExoMol-2020]_ species are available.

Absorption coefficient calculated with RADIS at 300 K, 1 atm are shown for all
[HITRAN-2016]_ molecules in the :ref:`HITRAN spectra page <label_examples_hitran_spectra>` .

- 1 	``'H2O'`` : 	Water 	(`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/0%20-%20H2O%20infrared%20spectrum.png>`__)
- 2 	``'CO2'`` : 	Carbon Dioxide    (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/1%20-%20CO2%20infrared%20spectrum.png>`__)
- 3 	``'O3'`` : 	Ozone  (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/2%20-%20O3%20infrared%20spectrum.png>`__)
- 4 	``'N2O'`` : 	Nitrogen oxide 	  (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/3%20-%20N2O%20infrared%20spectrum.png>`__)
- 5 	``'CO'`` : 	Carbon Monoxide    (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/4%20-%20CO%20infrared%20spectrum.png>`__)
- 6 	``'CH4'`` : 	Methane   (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/5%20-%20CH4%20infrared%20spectrum.png>`__)
- 7 	``'O2'`` : 	Oxygen
- 8 	``'NO'`` : 	Nitric Oxide   (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/7%20-%20NO%20infrared%20spectrum.png>`__)
- 9 	``'SO2'`` : 	Sulfur Dioxide    (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/8%20-%20SO2%20infrared%20spectrum.png>`__)
- 10 	``'NO2'`` : 	Nitrogen Dioxide     (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/9%20-%20NO2%20infrared%20spectrum.png>`__)
- 11 	``'NH3'`` : 	Ammonia  (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/10%20-%20NH3%20infrared%20spectrum.png>`__)
- 12 	``'HNO3'`` : 	Nitric Acid     (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/11%20-%20HNO3%20infrared%20spectrum.png>`__)
- 13 	``'OH'`` : 	Hydroxyl  (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/12%20-%20OH%20infrared%20spectrum.png>`__)
- 14 	``'HF'`` : 	Hydrogen Fluoride     (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/13%20-%20HF%20infrared%20spectrum.png>`__)
- 15 	``'HCl'`` : 	Hydrogen Chloride    (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/14%20-%20HCl%20infrared%20spectrum.png>`__)
- 16 	``'HBr'`` : 	Hydrogen Bromide     (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/15%20-%20HBr%20infrared%20spectrum.png>`__)
- 17 	``'HI'`` : 	Hydrogen Iodide   (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/16%20-%20HI%20infrared%20spectrum.png>`__)
- 18 	``'ClO'`` : 	Chlorine Monoxide    (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/17%20-%20ClO%20infrared%20spectrum.png>`__)
- 19 	``'OCS'`` : 	Carbonyl Sulfide     (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/18%20-%20OCS%20infrared%20spectrum.png>`__)
- 20 	``'H2CO'`` : 	Formaldehyde    (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/19%20-%20H2CO%20infrared%20spectrum.png>`__)
- 21 	``'HOCl'`` : 	Hypochlorous Acid   (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/20%20-%20HOCl%20infrared%20spectrum.png>`__)
- 22 	``'N2'`` : 	Nitrogen
- 23 	``'HCN'`` : 	Hydrogen Cyanide
- 24 	``'CH3Cl'`` : 	Methyl Chloride    (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/23%20-%20CH3Cl%20infrared%20spectrum.png>`__)
- 25 	``'H2O2'`` : 	Hydrogen Peroxide   (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/24%20-%20H2O2%20infrared%20spectrum.png>`__)
- 26 	``'C2H2'`` : 	Acetylene   (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/25%20-%20C2H2%20infrared%20spectrum.png>`__)
- 27 	``'C2H6'`` : 	Ethane  (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/26%20-%20C2H6%20infrared%20spectrum.png>`__)
- 28 	``'PH3'`` : 	Phosphine    (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/27%20-%20PH3%20infrared%20spectrum.png>`__)
- 29 	``'COF2'`` : 	Carbonyl Fluoride   (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/28%20-%20COF2%20infrared%20spectrum.png>`__)
- 30 	``'SF6'`` : 	Sulfur Hexafluoride
- 31 	``'H2S'`` : 	Hydrogen Sulfide     (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/30%20-%20H2S%20infrared%20spectrum.png>`__)
- 32 	``'HCOOH'`` : 	Formic Acid    (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/31%20-%20HCOOH%20infrared%20spectrum.png>`__)
- 33 	``'HO2'`` : 	Hydroperoxyl     (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/32%20-%20HO2%20infrared%20spectrum.png>`__)
- 34 	``'O'`` : 	Oxygen Atom
- 35 	``'ClONO2'`` : 	Chlorine Nitrate
- 36 	``'NO+'`` : 	Nitric Oxide Cation  (`spectrum <https://raw.githubusercontent.com/radis/radis-examples/master/hitran_spectra/out/35%20-%20NO%2B%20infrared%20spectrum.png>`__)
- 37 	``'HOBr'`` : 	Hypobromous Acid
- 38 	``'C2H4'`` : 	Ethylene
- 39 	``'CH3OH'`` : 	Methanol
- 40 	``'CH3Br'`` : 	Methyl Bromide
- 41 	``'CH3CN'`` : 	Acetonitrile
- 42 	``'CF4'`` : 	CFC-14
- 43 	``'C4H2'`` : 	Diacetylene
- 44 	``'HC3N'`` : 	Cyanoacetylene
- 45 	``'H2'`` : 	Hydrogen
- 46 	``'CS'`` : 	Carbon Monosulfide
- 47 	``'SO3'`` : 	Sulfur trioxide
- 48 	``'C2N2'`` : 	Cyanogen
- 49 	``'COCl2'`` : 	Phosgene

The code to calculate these spectra is also available on the :ref:`example page <label_examples_hitran_spectra>`.

See Also
--------

:py:data:`~radis.db.molecules.Molecules`,
:py:func:`~radis.db.molecules.getMolecule`

"""

MOLECULES_LIST_NONEQUILIBRIUM = (
    __supported_molecules_nonequilibrium__
) = _get_supported_molecules_nonequilibrium()
""" list: molecules that can be calculated in RADIS at nonequilibrium.
Spectroscopic constants to calculate energy levels are needed.

RADIS features some built-in :ref:`spectroscopic constants <label_db_spectroscopic_constants>`
for the following species ([HITRAN-2016]_ nomenclature):

- 2 	``'CO2'`` : 	Carbon Dioxide
- 5 	``'CO'`` : 	Carbon Monoxide


See Also
--------

:py:data:`~radis.db.molecules.Molecules`,
:py:func:`~radis.db.molecules.getMolecule`
"""


__all__ = [
    "MOLECULES_LIST_EQUILIBRIUM",
    "MOLECULES_LIST_NONEQUILIBRIUM",
    "Molecules",
    "getMolecule",
]



# -*- coding: utf-8 -*-
"""
Summary
-------

HITRAN database parser


Routine Listing
---------------

- :func:`~radis.io.hitran.hit2df`
- :func:`~radis.io.hitran.parse_local_quanta`
- :func:`~radis.io.hitran.parse_global_quanta`


-------------------------------------------------------------------------------


"""


import sys

# from radis.test.utils import getTestFile
import time
from collections import OrderedDict
from os.path import exists, getmtime

import pandas as pd
from numpy import int64

import radis
from radis.db.classes import (  # get_molecule_identifier,
    HITRAN_CLASS1,
    HITRAN_CLASS2,
    HITRAN_CLASS3,
    HITRAN_CLASS4,
    HITRAN_CLASS5,
    HITRAN_CLASS6,
    HITRAN_CLASS7,
    HITRAN_CLASS8,
    HITRAN_CLASS9,
    HITRAN_CLASS10,
    HITRAN_GROUP1,
    HITRAN_GROUP2,
    HITRAN_GROUP3,
    HITRAN_GROUP4,
    HITRAN_GROUP5,
    HITRAN_GROUP6,
    get_molecule,
)
from radis.io.cache_files import cache_file_name, load_h5_cache_file, save_to_hdf
from radis.io.tools import (
    drop_object_format_columns,
    parse_hitran_file,
    replace_PQR_with_m101,
)

# %% Parsing functions

# General case : HITRAN 2004
# fmt: off
columns_2004 = OrderedDict(
    [
        # name    # format # type  # description                                 # unit
        ("id", ("a2", int, "Molecular number", "")),
        ("iso", ("a1", int, "isotope number", "")),
        ("wav", ("a12", float, "vacuum wavenumber", "cm-1")),
        ("int", ("a10", float, "intensity at 296K", "cm-1/(molecule/cm-2)")),
        ("A", ("a10", float, "Einstein A coefficient", "s-1")),
        ("airbrd", ("a5", float, "air-broadened half-width at 296K", "cm-1.atm-1")),
        ("selbrd", ("a5", float, "self-broadened half-width at 296K", "cm-1.atm-1")),
        ("El", ("a10", float, "lower-state energy", "cm-1")),
        ("Tdpair", ("a4", float, "temperature-dependance exponent for Gamma air", "")),
        ("Pshft", ("a8", float, "air pressure-induced line shift at 296K", "cm-1.atm-1")),
        ("globu", ("a15", str, "electronic and vibrational global upper quanta", "")),
        ("globl", ("a15", str, "electronic and vibrational global lower quanta", "")),
        ("locu", ("a15", str, "electronic and vibrational local upper quanta", "")),
        ("locl", ("a15", str, "electronic and vibrational local lower quanta", "")),
        ("ierr", ("a6", str, "ordered list of indices corresponding to uncertainty estimates of transition parameters", "")),
        ("iref", ("a12", str, "ordered list of reference identifiers for transition parameters", "")),
        ("lmix", ("a1", str, "flag indicating the presence of additional data and code relating to line-mixing", "")),
        ("gp", ("a7", float, "upper state degeneracy", "")),
        ("gpp", ("a7", float, "lower state degeneracy", "")),
    ]
)
""" OrderedDict: parsing order of HITRAN 2004 format """
# fmt: on


def cast_to_int64_with_missing_values(dg, keys):
    """replace missing values of int64 columns with -1"""
    for c in keys:
        if dg.dtypes[c] != int64:
            dg[c] = dg[c].fillna(-1).astype(int64)


def hit2df(
    fname,
    cache=True,
    verbose=True,
    drop_non_numeric=True,
    load_wavenum_min=None,
    load_wavenum_max=None,
):
    """Convert a HITRAN/HITEMP [1]_ file to a Pandas dataframe

    Parameters
    ----------
    fname: str
        HITRAN-HITEMP file name
    cache: boolean, or ``'regen'`` or ``'force'``
        if ``True``, a pandas-readable HDF5 file is generated on first access,
        and later used. This saves on the datatype cast and conversion and
        improves performances a lot (but changes in the database are not
        taken into account). If False, no database is used. If ``'regen'``, temp
        file are reconstructed. Default ``True``.

    Other Parameters
    ----------------
    drop_non_numeric: boolean
        if ``True``, non numeric columns are dropped. This improves performances,
        but make sure all the columns you need are converted to numeric formats
        before hand. Default ``True``. Note that if a cache file is loaded it
        will be left untouched.
    load_wavenum_min, load_wavenum_max: float
        if not ``'None'``, only load the cached file if it contains data for
        wavenumbers above/below the specified value. See :py:func`~radis.io.cache_files.load_h5_cache_file`.
        Default ``'None'``.

    Returns
    -------
    df: pandas Dataframe
        dataframe containing all lines and parameters


    References
    ----------

    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__



    Notes
    -----

    Performances: see CDSD-HITEMP parser


    See Also
    --------

    :func:`~radis.io.cdsd.cdsd2df`
    """
    metadata = {}
    # Last modification time of the original file :
    metadata["last_modification"] = time.ctime(getmtime(fname))
    if verbose >= 2:
        print("Opening file {0} (cache={1})".format(fname, cache))
        print("Last modification time: {0}".format(metadata["last_modification"]))
    if load_wavenum_min and load_wavenum_max:
        assert load_wavenum_min < load_wavenum_max

    columns = columns_2004

    # Use cache file if possible
    fcache = cache_file_name(fname)
    if cache and exists(fcache):
        relevant_if_metadata_above = (
            {"wavenum_max": load_wavenum_min} if load_wavenum_min else {}
        )  # not relevant if wavenum_max of file is < wavenum min required
        relevant_if_metadata_below = (
            {"wavenum_min": load_wavenum_max} if load_wavenum_max else {}
        )  # not relevant if wavenum_min of file is > wavenum max required
        df = load_h5_cache_file(
            fcache,
            cache,
            valid_if_metadata_is=metadata,
            relevant_if_metadata_above=relevant_if_metadata_above,
            relevant_if_metadata_below=relevant_if_metadata_below,
            current_version=radis.__version__,
            last_compatible_version=radis.config["OLDEST_COMPATIBLE_VERSION"],
            verbose=verbose,
        )
        if df is not None:
            return df

    # Detect the molecule by reading the start of the file
    try:
        with open(fname) as f:
            mol = get_molecule(int(f.read(2)))
    except UnicodeDecodeError as err:
        raise ValueError(
            "You're trying to read a binary file {0} ".format(fname)
            + "instead of an HITRAN file"
        ) from err

    # %% Start reading the full file

    df = parse_hitran_file(fname, columns)

    # %% Post processing

    # assert one molecule per database only. Else the groupbase data reading
    # above doesnt make sense
    nmol = len(set(df["id"]))
    if nmol == 0:
        raise ValueError("Databank looks empty")
    elif nmol != 1:
        # Crash, give explicity error messages
        try:
            secondline = df.iloc[1]
        except IndexError:
            secondline = ""
        raise ValueError(
            "Multiple molecules in database ({0}). Current ".format(nmol)
            + "spectral code only computes 1 species at the time. Use MergeSlabs. "
            + "Verify the parsing was correct by looking at the first row below: "
            + "\n{0}".format(df.iloc[0])
            + "\n----------------\nand the second row "
            + "below: \n{0}".format(secondline)
        )

    # Add local quanta attributes, based on the HITRAN group
    try:
        df = parse_local_quanta(df, mol)
    except ValueError as err:
        # Empty strings (unlabelled lines) have been reported for HITEMP2010-H2O.
        # In this case, do not parse (makes non-equilibrium calculations impossible).
        # see https://github.com/radis/radis/issues/211
        if verbose:
            print(str(err))
            print("-" * 10)
            print(
                f"Impossible to parse local quanta in {fname}, probably an unlabelled line. Ignoring, but nonequilibrium calculations will not be possible. See details above."
            )

    # Add global quanta attributes, based on the HITRAN class
    try:
        df = parse_global_quanta(df, mol)
    except ValueError as err:
        # Empty strings (unlabelled lines) have been reported for HITEMP2010-H2O.
        # In this case, do not parse (makes non-equilibrium calculations impossible).
        # see https://github.com/radis/radis/issues/211
        if verbose:
            print(str(err))
            print("-" * 10)
            print(
                f"Impossible to parse global quanta in {fname}, probably an unlabelled line. Ignoring, but nonequilibrium calculations will not be possible. See details above."
            )

    # Remove non numerical attributes
    if drop_non_numeric:
        if "branch" in df:
            replace_PQR_with_m101(df)
        df = drop_object_format_columns(df, verbose=verbose)

    # cached file mode but cached file doesn't exist yet (else we had returned)
    if cache:
        new_metadata = {
            # Last modification time of the original file :
            "last_modification": time.ctime(getmtime(fname)),
            "wavenum_min": df.wav.min(),
            "wavenum_max": df.wav.max(),
        }
        if verbose:
            print(
                "Generating cache file {0} with metadata :\n{1}".format(
                    fcache, new_metadata
                )
            )
        try:
            save_to_hdf(
                df,
                fcache,
                metadata=new_metadata,
                version=radis.__version__,
                key="df",
                overwrite=True,
                verbose=verbose,
            )
        except PermissionError:
            if verbose:
                print(sys.exc_info())
                print("An error occured in cache file generation. Lookup access rights")
            pass

    # TODO : get only wavenum above/below 'load_wavenum_min', 'load_wavenum_max'
    # by parsing df.wav.   Completely irrelevant files are discarded in 'load_h5_cache_file'
    # but files that have partly relevant lines are fully loaded.
    # Note : cache file is generated with the full line list.

    return df


# %% Hitran global quanta classes


def _parse_HITRAN_class1(df):
    r"""Diatomic molecules: CO, HF, HCl, HBr, HI, N2, NO+


    Parameters
    ----------
    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----
    HITRAN syntax [1]_ :

    >>>       v
    >>>  13x I2

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    """

    # 1. Parse
    dgu = df["globu"].astype(str).str.extract(r"[ ]{13}(?P<vu>[\d ]{2})", expand=True)
    dgl = df["globl"].astype(str).str.extract(r"[ ]{13}(?P<vl>[\d ]{2})", expand=True)

    # 2. Convert to numeric
    cast_to_int64_with_missing_values(dgu, ["vu"])
    cast_to_int64_with_missing_values(dgl, ["vl"])

    # 3. Clean
    del df["globu"]
    del df["globl"]

    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class2(df):
    r"""Diatomic molecules with different electronic levels: O2


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_ :

    >>>      X   v
    >>>  12x A1 I2

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    """
    print("parse_global_quanta not implemented for molecules of HITRAN class 2")
    return df


def _parse_HITRAN_class3(df):
    r"""Diatomic molecules with doublet-Pi electronic state: NO, OH, ClO


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -------

    HITRAN syntax [1]_:

    >>>      X i     v1
    >>>  7x A1 A3 2x I2

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    """
    print("parse_global_quanta not implemented for molecules of HITRAN class 3")
    return df


def _parse_HITRAN_class4(df):
    r"""Parse linear triatomic class in HITRAN [1]_: N2O, OCS, HCN

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>     v1 v2 l2 v3
    >>>  7x I2 I2 I2 I2

    Note: I2 in regexp: [\d ]{2}

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    """

    # 1. Parse
    dgu = (
        df["globu"]
        .astype(str)
        .str.extract(
            r"[ ]{7}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})",
            expand=True,
        )
    )
    dgl = (
        df["globl"]
        .astype(str)
        .str.extract(
            r"[ ]{7}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})",
            expand=True,
        )
    )

    # 2. Convert to numeric
    cast_to_int64_with_missing_values(dgu, ["v1u", "v2u", "l2u", "v3u"])
    cast_to_int64_with_missing_values(dgl, ["v1l", "v2l", "l2l", "v3l"])

    # 3. Clean
    del df["globu"]
    del df["globl"]

    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class5(df):
    r"""Parse linear triatomic with large Fermi resonance in HITRAN [1]_: CO2

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>     v1 v2 l2 v3 r
    >>>  6x I2 I2 I2 I2 I1

    Note: I2 in regexp: [\d ]{2}

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    """

    # 1. Parse
    dgu = (
        df["globu"]
        .astype(str)
        .str.extract(
            r"[ ]{6}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})(?P<ru>\d)",
            expand=True,
        )
    )
    dgl = (
        df["globl"]
        .astype(str)
        .str.extract(
            r"[ ]{6}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})(?P<rl>\d)",
            expand=True,
        )
    )

    # 2. Convert to numeric
    cast_to_int64_with_missing_values(dgu, ["v1u", "v2u", "l2u", "v3u", "ru"])
    cast_to_int64_with_missing_values(dgl, ["v1l", "v2l", "l2l", "v3l", "rl"])

    # 3. Clean
    del df["globu"]
    del df["globl"]

    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class6(df):
    r"""Parse non-linear triatomic in HITRAN [1]_: H2O, O3, SO2, NO2, HOCl, H2S, HO2, HOBr

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>     v1 v2 v3
    >>>  9x I2 I2 I2

    Note: I2 in regexp: [\d ]{2}

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    """

    # 1. Parse
    dgu = (
        df["globu"]
        .astype(str)
        .str.extract(
            #        '[ ]{9}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<v3u>[\d ]{2})',
            r"[ ]{9}(?P<v1u>[\-\d ]{2})(?P<v2u>[\-\d ]{2})(?P<v3u>[\-\d ]{2})",
            expand=True,
        )
    )
    dgl = (
        df["globl"]
        .astype(str)
        .str.extract(
            #        '[ ]{9}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<v3l>[\d ]{2})',
            r"[ ]{9}(?P<v1l>[\-\d ]{2})(?P<v2l>[\-\d ]{2})(?P<v3l>[\-\d ]{2})",
            expand=True,
        )
    )
    # ... note @EP: in HITRAN H2O files, for iso=2, vibrational levels are
    # ... somehow negative. The regex above is adapted to catch negation signs with \-

    # 2. Convert to numeric
    cast_to_int64_with_missing_values(
        dgu,
        [
            "v1u",
            "v2u",
            "v3u",
        ],
    )
    cast_to_int64_with_missing_values(dgl, ["v1l", "v2l", "v3l"])

    # 3. Clean
    del df["globu"]
    del df["globl"]

    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_class7(df):
    r"""Parse linear tetratomic in HITRAN [1]_: C2H2

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database

    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    """
    print("parse_global_quanta not implemented for molecules of HITRAN class 7")
    return df


def _parse_HITRAN_class8(df):
    r"""Pyramidal tetratomic in HITRAN [1]_: NH3, PH3


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    """
    print("parse_global_quanta not implemented for molecules of HITRAN class 8")
    return df


def _parse_HITRAN_class9(df):
    r"""Non-linear tetratomic in HITRAN [1]_: H2CO, H2O2, COF2


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    """
    print("parse_global_quanta not implemented for molecules of HITRAN class 9")
    return df


def _parse_HITRAN_class10(df):
    r"""Pentatomic or greater polyatomic in HITRAN [1]_


    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax:

    >>>

    References
    ----------

    .. [1] `Table 3 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__

    """
    print("parse_global_quanta not implemented for molecules of HITRAN class 10")
    return df


# %% HITRAN Local quanta


def _parse_HITRAN_group1(df):
    r"""Parse asymmetric rotors (:py:attr:`~radis.db.classes.HITRAN_GROUP1` ):
    H2O, O3, SO2, NO2, HNO3, H2CO, HOCl, H2O2, COF2, H2S, HO2, HCOOH, ClONO2, HOBr, C2H4

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax: [1]_


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    """

    # 1. Parse

    # Ref [1] : locu
    # --------------
    # J'  | Ka' | Kc' | F'  | Sym'
    # I3  | I3  | I3  | A5  | A1
    dgu = (
        df["locu"]
        .astype(str)
        .str.extract(
            r"(?P<ju>[\d ]{3})(?P<Kau>[\-\d ]{3})(?P<Kcu>[\-\d ]{3})(?P<Fu>.{5})(?P<symu>.)",
            expand=True,
        )
    )
    # Ref [1] : locl
    # --------------
    # J'' | Ka''| Kc''| F'' | Sym''
    # I3  | I3  | I3  | A5  | A1
    dgl = (
        df["locl"]
        .astype(str)
        .str.extract(
            r"(?P<jl>[\d ]{3})(?P<Kal>[\-\d ]{3})(?P<Kcl>[\-\d ]{3})(?P<Fl>.{5})(?P<syml>.)",
            expand=True,
        )
    )
    # ... note @EP: in HITRAN H2O files, for iso=2, the Kau, Kcu can somehow
    # ... be negative. The regex above is adapted to catch negation signs with \-

    # 2. Convert to numeric
    cast_to_int64_with_missing_values(dgu, ["ju", "Kau", "Kcu"])
    cast_to_int64_with_missing_values(dgl, ["jl", "Kal", "Kcl"])

    # 3. Clean
    del df["locu"]
    del df["locl"]

    return pd.concat([df, dgu, dgl], axis=1)


def _parse_HITRAN_group2(df):
    r"""Parse diatomic and linear molecules (:py:attr:`~radis.db.classes.HITRAN_GROUP2` ):
    CO2, N2O, CO, HF, HCl, HBr, HI, OCS, N2, HCN, C2H2, NO+

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax: [1]


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    """

    # 1. Parse

    # Ref [1] : locu
    # --------------
    #     | F'  |
    # 10X | A5  |
    dgu = df["locu"].astype(str).str.extract(r"[ ]{10}(?P<Fu>.{5})", expand=True)
    # Ref [1] : locl
    # --------------
    #     | Br  | J'' | Sym''| F'' |
    # 5X  | A1  | I3  | A1   | A5  |
    dgl = (
        df["locl"]
        .astype(str)
        .str.extract(
            r"[ ]{5}(?P<branch>[\S]{1})(?P<jl>[\d ]{3})(?P<syml>.)(?P<Fl>.{5})",
            expand=True,
        )
    )

    # 2. Convert to numeric

    # dgl['jl'] = dgl.jl.apply(pd.to_numeric)
    cast_to_int64_with_missing_values(dgl, ["jl"])

    # 3. Clean
    del df["locu"]
    del df["locl"]

    return pd.concat([df, dgu, dgl], axis=1)


#                # 10X included in Fu
#               'Fu',     ('a15',  str,  'upper state total angular momentum including nuclear spin'  ,''         )),(
#               #'locl',   ('a15',  str,   'electronic and vibrational local lower quanta'  ,''                      )),(
#               'branch', ('a6',   str,     'O, P, Q, R, S branch symbol'                  ,''                      )),(
#               'jl',     ('a3',   int,    'lower state rotational quantum number'         ,''                      )),(
#               'sym',    ('a1',   str,     'symmetry'                                     ,''                      )),(
#               'Fl',     ('a5',   str,   'lower state total angular momentum including nuclear spin', ''         )),(


def _parse_HITRAN_group3(df):
    r"""Parse Spherical rotors (:py:attr:`~radis.db.classes.HITRAN_GROUP3` ) :
    SF6, CH4

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    """
    print("parse_local_quanta not implemented for molecules of HITRAN group 3")
    return df


def _parse_HITRAN_group4(df):
    r"""Parse symmetric rotors (:py:attr:`~radis.db.classes.HITRAN_GROUP4` ):
    CH3D, CH3Cl, C2H6, NH3, PH3, CH3OH

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    """
    print("parse_local_quanta not implemented for molecules of HITRAN group 4")
    return df


def _parse_HITRAN_group5(df):
    r"""Parse Triplet-Sigma ground electronic states (:py:attr:`~radis.db.classes.HITRAN_GROUP5` ) :
    O2

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    """
    print("parse_local_quanta not implemented for molecules of HITRAN group 5")
    return df


def _parse_HITRAN_group6(df):
    r"""Parse Doublet-Pi ground electronic states (:py:attr:`~radis.db.classes.HITRAN_GROUP6` ) :
    NO, OH, ClO

    Parameters
    ----------

    df: pandas Dataframe
        lines read from a HITRAN-like database


    Notes
    -----

    HITRAN syntax [1]_:


    References
    ----------

    .. [1] `Table 4 of Rothman et al. HITRAN 2004 <https://www.cfa.harvard.edu/hitran/Download/HITRAN04paper.pdf>`__


    """
    print("parse_local_quanta not implemented for molecules of HITRAN group 6")
    return df


# %% Reading function


def parse_local_quanta(df, mol):
    r"""
    Parameters
    ----------

    df: pandas Dataframe

    mol: str
        molecule name
    """

    if mol in HITRAN_GROUP1:
        df = _parse_HITRAN_group1(df)
    elif mol in HITRAN_GROUP2:
        df = _parse_HITRAN_group2(df)
    elif mol in HITRAN_GROUP3:
        df = _parse_HITRAN_group3(df)
    elif mol in HITRAN_GROUP4:
        df = _parse_HITRAN_group4(df)
    elif mol in HITRAN_GROUP5:
        df = _parse_HITRAN_group5(df)
    elif mol in HITRAN_GROUP6:
        df = _parse_HITRAN_group6(df)
    else:
        raise ValueError(
            "Unknown group for molecule {0}. Cant parse local quanta".format(mol)
        )

    return df


def parse_global_quanta(df, mol):
    r"""

    Parameters
    ----------

    df: pandas Dataframe

    mol: str
        molecule name
    """

    if mol in HITRAN_CLASS1:
        df = _parse_HITRAN_class1(df)
    elif mol in HITRAN_CLASS2:
        df = _parse_HITRAN_class2(df)
    elif mol in HITRAN_CLASS3:
        df = _parse_HITRAN_class3(df)
    elif mol in HITRAN_CLASS4:
        df = _parse_HITRAN_class4(df)
    elif mol in HITRAN_CLASS5:
        df = _parse_HITRAN_class5(df)
    elif mol in HITRAN_CLASS6:
        df = _parse_HITRAN_class6(df)
    elif mol in HITRAN_CLASS7:
        df = _parse_HITRAN_class7(df)
    elif mol in HITRAN_CLASS8:
        df = _parse_HITRAN_class8(df)
    elif mol in HITRAN_CLASS9:
        df = _parse_HITRAN_class9(df)
    elif mol in HITRAN_CLASS10:
        df = _parse_HITRAN_class10(df)
    else:
        raise ValueError(
            "Unknown class for molecule {0}. Cant parse global quanta".format(mol)
        )

    return df


# ======================================================
# %% Test


if __name__ == "__main__":

    from radis.test.test_io import _run_testcases

    print("Testing HITRAN parsing: ", _run_testcases())



# -*- coding: utf-8 -*-
"""
Test slit methods

Examples
--------

Run all tests::

    pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)::

    pytest -m fast


-------------------------------------------------------------------------------

"""


from os.path import basename
from warnings import catch_warnings, filterwarnings

import matplotlib.pyplot as plt
import numpy as np
import pytest
from numpy import abs, cos, linspace, pi, sqrt, tan

from radis.lbl.factory import SpectrumFactory
from radis.misc.arrays import nantrapz
from radis.misc.plot import fix_style, set_style
from radis.misc.printer import printm
from radis.phys.convert import dcm2dnm, dnm2dcm
from radis.phys.units import is_homogeneous

# from radis.lbl import SpectrumFactory
from radis.spectrum.models import calculated_spectrum, transmittance_spectrum
from radis.spectrum.spectrum import _cut_slices
from radis.test.utils import setup_test_line_databases
from radis.tools.database import load_spec
from radis.tools.slit import (
    get_effective_FWHM,
    get_FWHM,
    import_experimental_slit,
    offset_dilate_slit_function,
)

fig_prefix = basename(__file__) + ": "

# %% --------------------------------------------------------------------------
#  Test cases
# -----------------------------------------------------------------------------


def _clean(plot, close_plots):
    if plot:
        plt.ion()  # dont get stuck with Matplotlib if executing through pytest
        set_style()
        if close_plots:
            plt.close("all")


@pytest.mark.fast
def test_all_slit_shapes(
    FWHM=0.4, verbose=True, plot=True, close_plots=True, *args, **kwargs
):
    """Test all slit generation functions and make sure we get the expected FWHM"""

    _clean(plot, close_plots)

    # get spectrum
    from radis.spectrum.spectrum import Spectrum
    from radis.test.utils import getTestFile

    s = Spectrum.from_txt(
        getTestFile("calc_N2C_spectrum_Trot1200_Tvib3000.txt"),
        quantity="radiance_noslit",
        wunit="nm",
        unit="mW/cm2/sr/µm",
    )
    wstep = np.diff(s.get_wavelength())[0]

    # Plot all slits
    # ... gaussian
    s.apply_slit(FWHM, unit="nm", shape="gaussian", plot_slit=plot)
    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=2 * wstep)

    # ... triangular
    s.apply_slit(FWHM, unit="nm", shape="triangular", plot_slit=plot)
    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=2 * wstep)

    # ... trapezoidal
    s.apply_slit(
        (FWHM * 0.9, FWHM * 1.1), unit="nm", shape="trapezoidal", plot_slit=plot
    )
    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=2 * wstep)

    #    # ... trapezoidal
    #    s.apply_slit(FWHM, unit='nm', shape='trapezoidal', plot_slit=plot, norm='max')
    #    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=1.1*wstep)

    # ... experimental
    s.apply_slit(getTestFile("slitfunction.txt"), unit="nm", plot_slit=plot)
    assert np.isclose(get_effective_FWHM(*s.get_slit()), FWHM, atol=0.01)
    # note that we're applying a slit function measured at 632.5 nm to a Spectrum
    # at 4.7 µm. It's just good for testing the functions

    #    # ... experimental, convolve with max
    #    s.apply_slit(getTestFile('slitfunction.txt'), unit='nm', norm_by='max', plot_slit=plot)
    #    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=1.1*wstep)

    if verbose:
        print("\n>>> _test_all_slits yield correct FWHM (+- wstep) : OK\n")

    return True  # nothing defined yet


@pytest.mark.fast
def test_slit_unit_conversions_spectrum_in_cm(
    verbose=True, plot=True, close_plots=True, *args, **kwargs
):
    """Test that slit is consistently applied for different units

    Assert that:

    - calculated FWHM is the one that was applied

    """

    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec

    _clean(plot, close_plots)

    # %% Get a Spectrum (stored in cm-1)
    s_cm = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)

    s_cm.rescale_mole_fraction(1)  # just because it makes better units
    s_cm.update()
    wstep = s_cm.conditions["wstep"]

    assert s_cm.get_waveunit() == "cm-1"  # ensures it's stored in cm-1

    for shape in ["gaussian", "triangular"]:

        # Apply slit in cm-1
        slit_cm = 2
        s_cm.name = "Spec in cm-1, slit {0:.2f} cm-1".format(slit_cm)
        s_cm.apply_slit(slit_cm, unit="cm-1", shape=shape, mode="same")
        # ... mode=same to keep same output length. It helps compare both Spectra afterwards
        # in cm-1 as that's s.get_waveunit()
        fwhm = get_FWHM(*s_cm.get_slit())
        assert np.isclose(slit_cm, fwhm, atol=2 * wstep)

        # Apply slit in nm this time
        s_nm = s_cm.copy()
        w_cm = s_nm.get_wavenumber()
        slit_nm = dcm2dnm(slit_cm, w_cm[len(w_cm) // 2])
        s_nm.name = "Spec in cm-1, slit {0:.2f} nm".format(slit_nm)
        s_nm.apply_slit(slit_nm, unit="nm", shape=shape, mode="same")

        plotargs = {}
        if plot:
            plotargs["title"] = "test_slit_unit_conversions: {0} ({1} cm-1)".format(
                shape, slit_cm
            )
        s_cm.compare_with(
            s_nm,
            spectra_only="radiance",
            rtol=1e-3,
            verbose=verbose,
            plot=plot,
            **plotargs
        )


@pytest.mark.fast
def test_slit_unit_conversions_spectrum_in_nm(
    verbose=True, plot=True, close_plots=True, *args, **kwargs
):
    """Test that slit is consistently applied for different units

    Assert that:

    - calculated FWHM is the one that was applied

    """

    from radis.spectrum.spectrum import Spectrum
    from radis.test.utils import getTestFile

    _clean(plot, close_plots)

    # %% Get a Spectrum (stored in nm)

    s_nm = Spectrum.from_txt(
        getTestFile("calc_N2C_spectrum_Trot1200_Tvib3000.txt"),
        quantity="radiance_noslit",
        wunit="nm",
        unit="mW/cm2/sr/µm",
        conditions={"self_absorption": False},
    )

    with catch_warnings():
        filterwarnings(
            "ignore", "Condition missing to know if spectrum is at equilibrium:"
        )
        # just because it makes better units
        s_nm.rescale_path_length(1, 0.001)

    wstep = np.diff(s_nm.get_wavelength())[0]

    assert s_nm.get_waveunit() == "nm"  # ensures it's stored in cm-1

    for shape in ["gaussian", "triangular"]:

        # Apply slit in nm
        slit_nm = 0.5
        s_nm.name = "Spec in nm, slit {0:.2f} nm".format(slit_nm)
        s_nm.apply_slit(slit_nm, unit="nm", shape=shape, mode="same")
        # ... mode=same to keep same output length. It helps compare both Spectra afterwards
        # in cm-1 as that's s.get_waveunit()
        fwhm = get_FWHM(*s_nm.get_slit())
        assert np.isclose(slit_nm, fwhm, atol=2 * wstep)

        # Apply slit in nm this time
        s_cm = s_nm.copy()
        w_nm = s_nm.get_wavelength()
        slit_cm = dnm2dcm(slit_nm, w_nm[len(w_nm) // 2])
        s_cm.name = "Spec in nm, slit {0:.2f} cm-1".format(slit_cm)
        s_cm.apply_slit(slit_cm, unit="cm-1", shape=shape, mode="same")

        plotargs = {}
        if plot:
            plotargs["title"] = "test_slit_unit_conversions: {0} ({1} nm)".format(
                shape, slit_nm
            )
        s_nm.compare_with(
            s_cm,
            spectra_only="radiance",
            rtol=1e-3,
            verbose=verbose,
            plot=plot,
            **plotargs
        )

    # %%


def test_convoluted_quantities_units(*args, **kwargs):
    """Test that units are correctly convoluted after convolution"""

    from radis.test.utils import getTestFile

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.5.spec"), binary=True)

    s.update(verbose=False)

    assert s.units["radiance_noslit"] == "mW/cm2/sr/nm"
    assert s.units["transmittance_noslit"] == ""

    s.apply_slit(0.5, norm_by="area", verbose=False)

    assert s.units["radiance"] == "mW/cm2/sr/nm"
    assert s.units["transmittance"] == ""

    s.apply_slit(0.5, norm_by="max", verbose=False)

    assert is_homogeneous(s.units["radiance"], "mW/cm2/sr")
    assert s.units["transmittance"] == "nm"  # whatever that means


@pytest.mark.fast
def test_against_specair_convolution(
    plot=True, close_plots=True, verbose=True, debug=False, *args, **kwargs
):

    _clean(plot, close_plots)

    # Test
    from radis.test.utils import getTestFile

    # Plot calculated vs convolved with slit
    # Specair units: mW/cm2/sr/µm
    w, I = np.loadtxt(getTestFile("calc_N2C_spectrum_Trot1200_Tvib3000.txt")).T
    s = calculated_spectrum(
        w, I, conditions={"Tvib": 3000, "Trot": 1200}, Iunit="mW/cm2/sr/µm"
    )

    if plot:
        fig = plt.figure(fig_prefix + "SPECAIR convoluted vs SPECAIR non convoluted")
        s.plot("radiance_noslit", nfig=fig.number, label="calc")
    slit_nm = 0.1
    s.apply_slit(slit_nm, shape="triangular", norm_by="area")
    if plot:
        s.plot(
            "radiance",
            nfig=fig.number,
            label="slit {0}nm".format(slit_nm),
            color="r",
            lw=2,
        )
        plt.legend()

    # Compare with Specair slit function
    s.apply_slit(slit_nm, norm_by="max")
    # Note unit conversion from mW/cm2/sr/um*nm is properly done!
    if plot:
        fig = plt.figure(fig_prefix + "convoluted RADIS vs convoluted SPECAIR")
        s.plot(
            "radiance",
            nfig=fig.number,
            label="slit {0}nm, norm with max".format(slit_nm),
            color="r",
            lw=2,
            Iunit="mW/cm2/sr",
        )
    # ... Get initial spectrum convoluted with Specair
    ws, Is = np.loadtxt(
        getTestFile("calc_N2C_spectrum_Trot1200_Tvib3000_slit0.1.txt")
    ).T  # Specair units: mW/cm2/sr
    if plot:
        plt.plot(ws, Is, "k", label="normalized in Specair (sides not cropped)")

    # ... Test output is the same
    wu, Iu = s.get("radiance", Iunit="mW/cm2/sr")

    As = nantrapz(Is, ws)  # Todo one day: replace with get_power() function
    Au = nantrapz(Iu, wu)
    if verbose:
        print(
            (
                "Integrals should match: {0:.2f} vs {1:.2f} ({2:.2f}% error)".format(
                    As, Au, 100 * abs(As - Au) / As
                )
            )
        )
    assert np.isclose(As, Au, rtol=1e-2)

    # Test resampling
    s.resample(linspace(376, 380.6, 3000), unit="nm")
    s.apply_slit(slit_nm, norm_by="max")
    if plot:
        s.plot(
            "radiance",
            Iunit="mW/cm2/sr",
            lw=3,
            nfig=fig.number,
            color="b",
            zorder=-3,
            label="Resampled",
        )
        plt.legend()

    if verbose:
        print("\n>>>Testing spectrum slit matches Specair: OK")

    return True


@pytest.mark.fast
def test_normalisation_mode(plot=True, close_plots=True, verbose=True, *args, **kwargs):
    """Test norm_by = 'area' vs norm_by = 'max'"""

    from radis.test.utils import getTestFile

    _clean(plot, close_plots)

    # %% Compare spectra convolved with area=1 and max=1
    # Slit in nm
    # Spectrum in nm
    # Specair units: mW/cm2/sr/µm
    w, I = np.loadtxt(getTestFile("calc_N2C_spectrum_Trot1200_Tvib3000.txt")).T
    s = calculated_spectrum(
        w, I, conditions={"Tvib": 3000, "Trot": 1200}, Iunit="mW/cm2/sr/µm"
    )

    FWHM = 2

    s.apply_slit(FWHM, norm_by="area")  # spectrum convolved with area=1
    w_area, I_area = s.get("radiance")
    if plot:
        fig = plt.figure(fig_prefix + "Spectrum in nm + slit in nm")
        fig.clear()
        ax = fig.gca()
        s.plot(nfig=fig.number, wunit="nm", label="norm_by: area", lw=3)
    s.apply_slit(FWHM, norm_by="max")  # spectrum convolved with max=1
    w_max, I_max = s.get("radiance", wunit="nm")
    if plot:
        ax.plot(w_max, I_max / FWHM, "r", label="(norm_by:max)/FWHM")
        ax.legend(loc="best")
    assert np.allclose(I_area, I_max / FWHM, equal_nan=True)
    if verbose:
        print("equivalence of normalisation mode for spectrum in 'nm': OK")

    # %% Compare spectra convolved with area=1 and max=1
    # Slit in nm
    # Spectrum in cm-1

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update()
    # spectrum convolved with area=1
    s.apply_slit(FWHM, norm_by="area", plot_slit=plot)
    w_area, I_area = s.get("radiance")
    if plot:
        fig = plt.figure(fig_prefix + "Spectrum in cm-1 + slit in nm")
        fig.clear()
        ax = fig.gca()
        s.plot(nfig=fig.number, wunit="nm", label="norm_by: area", lw=3)
    # spectrum convolved with max=1
    s.apply_slit(FWHM, norm_by="max", plot_slit=plot)
    w_max, I_max = s.get("radiance", wunit="nm")
    if plot:
        ax.plot(w_max, I_max / FWHM, "r", label="(norm_by:max)/FWHM")
        ax.legend(loc="best")
    assert np.allclose(I_area, I_max / FWHM, equal_nan=True)
    if verbose:
        print("equivalence of normalisation mode for spectrum in 'cm-1': {0}: OK")
    assert is_homogeneous(s.units["radiance"], "mW/cm2/sr")
    if verbose:
        print(
            (
                "radiance unit ({0}) is homogeneous to 'mW/cm2/sr': OK".format(
                    s.units["radiance"]
                )
            )
        )

    return True


@pytest.mark.fast
def test_slit_energy_conservation(
    verbose=True, plot=True, close_plots=True, *args, **kwargs
):
    """Convoluted and non convoluted quantities should have the same area
    (difference arises from side effects if the initial spectrum is not 0 on
    the sides"""

    from radis.test.utils import getTestFile

    _clean(plot, close_plots)

    if verbose:
        print("\n>>> _test_slit_energy_conservation\n")

    s = calculated_spectrum(
        *np.loadtxt(getTestFile("calc_N2C_spectrum_Trot1200_Tvib3000_slit0.1.txt")).T,
        wunit="nm",
        Iunit="mW/cm2/sr/nm"
    )  # arbitrary)

    P = s.get_power(unit="mW/cm2/sr")
    s.apply_slit(0.5, norm_by="area")
    w, I = s.get("radiance", wunit="nm", Iunit="mW/cm2/sr/nm")
    Pc = abs(np.trapz(I[~np.isnan(I)], x=w[~np.isnan(I)]))  # mW/cm2/sr

    b = np.isclose(P, Pc, 3e-2)

    if plot:
        fig = plt.figure(fig_prefix + "energy conservation during resampling")
        s.plot(nfig=fig.number, label="{0:.1f} mW/cm2/sr".format(P))
        s.plot("radiance_noslit", nfig=fig.number, label="{0:.1f} mW/cm2/sr".format(Pc))
        plt.title("Energy conservation: {0}".format(b))
        plt.legend()
        plt.tight_layout()

    assert np.isclose(P, Pc, 3e-2)

    return True


# Function used to test Slit dispersion


def dirac(w0, width=20, wstep=0.009):
    """Return a Dirac in w0. Plus some space on the side"""

    w_side_p = np.arange(w0, w0 + width, wstep)
    w_side_n = np.arange(w0, w0 - width, -wstep)
    w = np.hstack((w_side_n[:0:-1], w_side_p))
    I = np.zeros_like(w)
    I[len(I) // 2] = 1 / wstep

    return w, I


def linear_dispersion(w, f=750, phi=-6, m=1, gr=300):
    """dlambda / dx
    Default values correspond to Acton 750i

    Parameters
    ----------

    f: focal length (mm)
         default 750 (SpectraPro 2750i)

    phi: angle in degrees (°)
        default 9

    m: order of dispersion
        default 1

    gr: grooves spacing (gr/mm)
        default 300
    """
    # correct units:
    phi *= 2 * pi / 360
    d = 1e-3 / gr
    disp = w / (2 * f) * (tan(phi) + sqrt((2 * d / m / (w * 1e-9) * cos(phi)) ** 2 - 1))
    return disp  # to nm/mm


@pytest.mark.fast
def test_linear_dispersion_effect(
    verbose=True, plot=True, close_plots=True, *args, **kwargs
):
    """A test case to show the effect of wavelength dispersion (cf spectrometer
    reciprocal function) on the slit function

    Test succeeds if a :py:data:`~radis.misc.warning.SlitDispersionWarning`
    is correctly triggered and if the FWHM is properly dilated.
    """

    from radis.test.utils import getTestFile

    _clean(plot, close_plots)

    w_slit, I_slit = import_experimental_slit(getTestFile("slitfunction.txt"))

    if plot:
        plt.figure(fig_prefix + "Linear dispersion effect")
        plt.plot(
            w_slit,
            I_slit,
            "--k",
            label="Exp: FWHM @{0}nm: {1:.3f} nm".format(
                632.6, get_effective_FWHM(w_slit, I_slit)
            ),
        )

    from radis.misc.warning import SlitDispersionWarning

    with pytest.warns(
        SlitDispersionWarning
    ):  # expect a "large slit dispersion" warning

        # Test how slit function FWHM scales with linear_dispersion
        for w0, FWHM in zip([380, 1000, 4200, 5500], [0.393, 0.385, 0.280, 0.187]):
            w, I = dirac(w0)

            wc, Ic = offset_dilate_slit_function(
                w_slit,
                I_slit,
                w,
                slit_dispersion=linear_dispersion,
                threshold=0.01,
                verbose=False,
            )
            assert np.isclose(FWHM, get_effective_FWHM(wc, Ic), atol=0.001)
            if plot:
                plt.plot(
                    wc,
                    Ic,
                    label="FWHM @{0:.2f} nm: {1:.3f} nm".format(
                        w0, get_effective_FWHM(wc, Ic)
                    ),
                )

    if plot:
        plt.xlabel("Wavelength (nm)")
        plt.ylabel("Dirac $x$ slit function")
        plt.legend(loc="best", prop={"size": 15})
        fix_style()

    return True


@pytest.mark.fast
def test_cut_slices(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    """A test case to verify that _cut_slices does cut the spectrum into slices

    Test fails if a :py:data:`~radis.misc.warning.SlitDispersionWarning`
    is  triggered and succeeds if the spectrum is sliced into 8 segments.
    """
    from radis.misc.warning import SlitDispersionWarning

    _clean(plot, close_plots)

    threshold = 0.01
    w = np.arange(4000, 4400, 0.01)
    w_slit = np.arange(4198, 4202, 0.1)

    slices = _cut_slices(w, w_slit, linear_dispersion, threshold)

    for sl in slices:
        try:
            offset_dilate_slit_function(
                w_slit, np.ones_like(w_slit), w[sl], linear_dispersion, threshold, True
            )
        except SlitDispersionWarning:
            return False
        if plot:
            plt.plot(
                w,
                sl,
                label="Slice {0:.2f}-{1:.2f} nm , slit dispersion ratio: {2:.3f}".format(
                    w[sl][0],
                    w[sl][-1],
                    linear_dispersion(w[sl][-1]) / linear_dispersion(w[sl][0]),
                ),
            )
    if plot:
        plt.title("Cut slices, threshold (boundaries removed) = {0}".format(threshold))
        plt.xlabel("Wavelength (nm)")
        plt.ylabel("Slices (Boolean)")
        plt.legend(loc="best", prop={"size": 15})

    assert len(slices) == 8
    slices = _cut_slices(w[::-1], w_slit, linear_dispersion, threshold)
    assert len(slices) == 8

    return True


@pytest.mark.fast
def test_auto_correct_dispersion(
    verbose=True, plot=True, close_plots=True, *args, **kwargs
):
    """A test case to show the effect of wavelength dispersion (cf spectrometer
    reciprocal function) on the slit function

    Test succeeds if there is a difference between convoluted spectra (with and without slit dispersion)
    """

    from radis.test.utils import getTestFile

    _clean(plot, close_plots)

    slit_measured_632nm = getTestFile("slitfunction.txt")

    w, I = np.loadtxt(getTestFile("calc_N2C_spectrum_Trot1200_Tvib3000.txt")).T
    s = calculated_spectrum(
        w, I, conditions={"Tvib": 3000, "Trot": 1200}, Iunit="mW/cm2/sr/µm"
    )
    s2 = s.copy()

    def slit_dispersion(w):
        return linear_dispersion(w, f=750, phi=-6, m=1, gr=2400)

    s.apply_slit(slit_measured_632nm)

    if plot:
        w_slit_632, I_slit_632 = import_experimental_slit(
            getTestFile("slitfunction.txt")
        )
        w_full_range = np.linspace(w.min(), w_slit_632.max())
        plt.figure(
            "Spectrometer Dispersion (f={0}mm, phi={1}°, gr={2}".format(750, -0.6, 2400)
        )
        plt.plot(w_full_range, slit_dispersion(w_full_range))
        plt.xlabel("Wavelength (nm)")
        plt.ylabel("Reciprocal Linear Dispersion")

        # Compare 2 spectra
        s.plot(nfig="Linear dispersion effect", color="r", label="not corrected")

    s2 = s2.apply_slit(
        slit_measured_632nm, slit_dispersion=slit_dispersion, inplace=False
    )

    if plot:
        s2.plot(nfig="same", color="k", label="corrected")
        plt.legend()
        # Plot different slits:
        s2.plot_slit()
    assert np.isclose(
        s.take("radiance").max() / (s2.take("radiance").max()), 1.183, atol=0.001
    )
    return True


@pytest.mark.fast
def test_resampling(rtol=1e-2, verbose=True, plot=True, warnings=True, *args, **kwargs):
    """Test what happens when a spectrum in nm or cm-1, is convolved
    with a slit function in nm. In particular, slit function is generated
    in the spectrum unit, and spectrum is resampled if not evenly spaced"""

    if verbose:
        printm("Test auto resampling")

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/radis.json if not there

    plCO = SpectrumFactory(
        wavenum_min=2230,
        wavenum_max=2260,
        mole_fraction=0.02,
        path_length=100,  # cm
        truncation=10,  # cm^-1
        wstep=0.02,
        isotope=[1, 2, 3],
        verbose=verbose,
    )
    plCO.warnings["MissingSelfBroadeningWarning"] = "ignore"
    plCO.load_databank("HITRAN-CO-TEST")
    sCO = plCO.eq_spectrum(Tgas=300)

    w_nm, T_nm = sCO.get("transmittance_noslit", wunit="nm")
    w_nm, I_nm = sCO.get("radiance_noslit", wunit="nm", Iunit="mW/cm2/sr/nm")
    sCO_nm = transmittance_spectrum(
        w_nm, T_nm, wunit="nm"
    )  # a new spectrum stored in nm
    # sCO_nm = theoretical_spectrum(w_nm, I_nm, wunit='nm', Iunit='mW/cm2/sr/nm') #  a new spectrum stored in nm

    if plot:
        fig = plt.figure(fig_prefix + "auto-resampling")
        sCO.plot(
            "transmittance_noslit",
            wunit="cm-1",
            nfig=fig.number,
            marker="o",
            color="k",
            lw=3,
            ms=10,
            label="(stored in cm-1)",
        )
        plt.title("No slit function")
        sCO_nm.plot(
            "transmittance_noslit",
            wunit="cm-1",
            nfig=fig.number,
            marker="o",
            color="r",
            label="(stored in nm)",
        )
        #            plt.xlim((2246.58, 2247.52))
        #            plt.ylim((0.87, 1.01))
        plt.legend()

    slit_function = 0.8
    slit_unit = "cm-1"
    sCO.apply_slit(slit_function, unit=slit_unit)
    sCO_nm.apply_slit(slit_function, unit=slit_unit)

    if plot:
        fig = plt.figure(fig_prefix + "auto-resampling (after convolution)")
        sCO.plot(
            "transmittance",
            wunit="cm-1",
            nfig=fig.number,
            marker="o",
            color="k",
            lw=3,
            ms=10,
            label="(stored in cm-1)",
        )
        plt.title("Slit function: {0} {1}".format(slit_function, slit_unit))
        sCO_nm.plot(
            "transmittance",
            wunit="cm-1",
            nfig=fig.number,
            marker="o",
            color="r",
            label="(stored in nm)",
        )

        #            plt.xlim((2246.58, 2247.52))
        #            plt.ylim((0.87, 1.01))
        plt.legend()

    w_conv, T_conv = sCO.get("transmittance", wunit="cm-1")
    w_nm_conv, T_nm_conv = sCO_nm.get("transmittance", wunit="cm-1")

    error = abs(
        (nantrapz(1 - T_conv, w_conv) - nantrapz(1 - T_nm_conv, w_nm_conv))
        / nantrapz(1 - T_nm_conv, w_nm_conv)
    )

    if verbose:
        printm("\n>>> _test_resampling\n")
    if verbose:
        printm(
            "Error between 2 spectra ({0:.2f}%) < {1:.2f}%: {2}".format(
                error * 100, rtol * 100, bool(error < rtol)
            )
        )
    assert bool(error < rtol)


def _run_testcases(plot=True, close_plots=False, verbose=True, *args, **kwargs):

    # Validation
    test_against_specair_convolution(
        plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs
    )

    # Different modes
    test_normalisation_mode(
        plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs
    )

    # Resampling
    test_slit_energy_conservation(
        plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs
    )

    # Slit dispersion
    test_linear_dispersion_effect(
        plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs
    )
    test_cut_slices(
        plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs
    )
    test_auto_correct_dispersion(
        plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs
    )

    # Different shapes
    test_all_slit_shapes(
        plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs
    )

    # Units
    test_slit_unit_conversions_spectrum_in_cm(
        verbose=verbose, plot=plot, close_plots=close_plots, *args, **kwargs
    )
    test_slit_unit_conversions_spectrum_in_nm(
        verbose=verbose, plot=plot, close_plots=close_plots, *args, **kwargs
    )
    test_convoluted_quantities_units(*args, **kwargs)

    test_resampling(plot=plot, verbose=verbose, *args, **kwargs)

    return True


if __name__ == "__main__":
    print(("Testing slit.py: ", _run_testcases(plot=True)))



# -*- coding: utf-8 -*-
"""
Created on Mon Jul 19 22:44:39 2021

@author: erwan
"""

import pytest

from radis.io.exomol import get_exomol_database_list, get_exomol_full_isotope_name


def test_exomol_parsing_functions(verbose=True, *args, **kwargs):
    """Test functions used to parse ExoMol website"""

    assert get_exomol_full_isotope_name("H2O", 1) == "1H2-16O"
    assert get_exomol_full_isotope_name("H2O", 4) == "1H-2H-16O"
    assert get_exomol_full_isotope_name("CO2", 3) == "16O-12C-18O"
    assert get_exomol_full_isotope_name("H2O", 1) == "1H2-16O"
    assert get_exomol_full_isotope_name("NH3", 1) == "14N-1H3"
    assert get_exomol_full_isotope_name("H2S", 1) == "1H2-32S"
    assert get_exomol_full_isotope_name("FeH", 1) == "56Fe-1H"
    assert get_exomol_full_isotope_name("CH4", 1) == "12C-1H4"
    assert get_exomol_full_isotope_name("C2H4", 1) == "12C2-1H4"
    assert get_exomol_full_isotope_name("C2H2", 1) == "12C2-1H2"

    #  Note : this may change if new databases are added ; test would have to
    # be updated in that case.
    databases, recommended = get_exomol_database_list("CH4", "12C-1H4")
    assert databases == ["xsec-YT10to10", "YT10to10", "YT34to10"]
    assert recommended == "YT34to10"

    # Test that these databases are found
    KNOWN_EXOMOL_DATABASE_NAMES = {
        "H2O": {"1H2-16O": ["POKAZATEL"]},
        "NH3": {"14N-1H3": ["CoYuTe"]},
        "H2S": {"1H2-32S": ["AYT2"]},
        "FeH": {"56Fe-1H": ["MoLLIST"]},
    }
    for molecule, val in KNOWN_EXOMOL_DATABASE_NAMES.items():
        for isotope_name, lookup_databases in val.items():
            known_databases, _ = get_exomol_database_list(molecule, isotope_name)
            for dat in lookup_databases:
                assert dat in known_databases


@pytest.mark.needs_connection
def test_calc_exomol_spectrum(verbose=True, plot=True, *args, **kwargs):
    """Auto-fetch and calculate a SiO spectrum from the ExoMol database

    https://github.com/radis/radis/pull/320#issuecomment-884508206
    """
    from radis import calc_spectrum

    s = calc_spectrum(
        1080,
        1320,  # cm-1
        molecule="SiO",
        isotope="1",
        pressure=1.01325,  # bar
        Tgas=1000,  # K
        mole_fraction=0.1,
        path_length=1,  # cm
        broadening_method="fft",  # @ dev: Doesn't work with 'voigt'
        databank="exomol",  # or use ('exomol', "EBJT") for a specific database ("EBJT")
        verbose=verbose,
    )
    s.apply_slit(1, "cm-1")  # simulate an experimental slit
    if plot:
        s.plot("radiance")


if __name__ == "__main__":
    test_exomol_parsing_functions()
    test_calc_exomol_spectrum()
